/*
 * Public domain
 * sys/wait.h compatibility shim
 */

#ifndef _MSC_VER
#include_next <sys/wait.h>
#else

#ifndef IKED_COMPAT_SYS_WAIT_H
#define IKED_COMPAT_SYS_WAIT_H

/* FIXME */

#endif /* !IKED_COMPAT_SYS_WAIT_H */

#endif
